! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module edetect_io_sens
  
  use errorHandling
  use param  
  use dal
  use cal
  use cal_aux       ! requires ssclib in DEPEND and ssc in src/Makefile
  use wcs_aux       ! -"-

  implicit none

  private

  public cp_head, get_files, co_bgima, co_expima, co_mask  !!! , co_image
  
! MAX_IMA              max number of images
! MAX_INS              max number of different instruments 
! MAX_BANDS            max number of energy bands

  integer, parameter, public :: max_ima = 54 
  integer, parameter, public :: max_ins =  9
  integer, parameter, public :: max_bands = 6

! NUM_INSTR            number of instruments
! NUM_IMAGE(MAX_INS)   number of images per instrumet
! NIM                  total number of images

  integer, public :: num_instr, nim  
  integer, dimension(max_ins), public :: num_image       

! IMAGE                input images
! BG_IMAGE             background images
! EXP_IMAGE            exposure images

!!!  integer, dimension(:,:,:), allocatable, public  :: iimage
  real, dimension (:,:,:), allocatable, public   :: bg_image, exp_image
  logical, dimension(:,:,:), allocatable, public :: mask_image

  integer, dimension(2), public                  :: npix

! IP(MAX_INS,MAX_BANDS)pointer field to manage images

  integer, dimension(max_ins, max_bands), public :: ip
  integer, dimension(max_ins), public            :: expip

  integer, public :: imxmin, imxmax, imymin, imymax, imusex, imusey
  real(kind=single), public :: xcen1, ycen1
  real(kind=double), public :: crpix1,crpix2,crval1,crval2, &
                               crdelt1,crdelt2,crota2
  real(kind=double), dimension(max_ins), public :: ra_pnt, dec_pnt, pa_pnt
  real(kind=single), dimension(3), public       :: xoff, yoff

!  real(kind=double) :: ra_pnt, dec_pnt, pa_pnt
!  real(kind=double) :: crpix1,crpix2,crdelt1,crdelt2,crota2

  type(WcsT), public :: wcs
  type(SpacecraftAttitudeType), dimension(max_ima), public :: attitudes
  real(kind=double),            dimension(max_ima), public :: timestamps
  

  real(kind=double) :: pi = 3.141592654

contains

  !---------------------------------------------------------------------
  subroutine cp_attr (set1, set2, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr

    integer i

    do i=1,size(attr)
      if (hasAttribute(set1,attr(i))) then
        call setAttribute(set2,attribute(set1,attr(i)))
      else 
        if (hasAttribute(arr,attr(i))) then
          call setAttribute(set2,attribute(arr,attr(i)))
        end if
      end if
    end do

  end subroutine cp_attr

  !---------------------------------------------------------------------
  subroutine cp_head (set1, set2, content, mod)
  !---------------------------------------------------------------------

    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT) :: arr, arr2
    character(len=*), intent (IN) :: content
    integer, intent(IN) :: mod

    arr = array (set1,"PRIMARY",read)

    !!! IT 2014/05
    !!! copy full dataset header and array header of images
    call setAttributes( attributable(set2), attributable(set1) )
    ! sensitivity maps are created without array -> merge headers
    call addAttributes( attributable(set2), attributable(arr) )

!!!    if (mod > 0) then
!!!      call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
!!!    end if
!!!
!!!    if (mod == 2) then
!!!      call cp_attr(set1,set2,arr,(/"BAND"/) )
!!!    end if
!!!
!!!    call cp_attr(set1,set2,arr, &
!!!    (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
!!!       "ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
!!!       "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
!!!       "CDELT2  ","CTYPE1  ","CTYPE2  ","SUBMODE ","RADECSYS","EQUINOX ", &
!!!       "EXPOSURE","DATE-OBS","DATE-END","OBSERVER","EXPIDSTR","ATT_SRC "/))

    call setAttribute(set2,"CONTENT", content, "Content of file")

    call release(arr)

  end subroutine cp_head

  !--------------------------------------------------------------------
  subroutine get_files (param_name, imag, instrume, band, first)
  !---------------------------------------------------------------------
 
    character(len=*), intent(IN) :: param_name
    logical, intent(IN) :: first
    character(len=1024), dimension(*), intent(OUT) :: imag
    character(len=21), dimension(*), intent(OUT) :: instrume  
    integer, dimension(*), intent(OUT) :: band

    integer :: i, j, ib, ban
    integer :: nimage, ipoint
!!!    character(len=7) :: teles, instru, expid  
    character(len=3) :: teles
    character(len=5) :: instru
    character(len=13) :: expid
    character(len=16) :: ins  
    logical, dimension(max_ima) :: double

    type(DataSetT) :: set
    type(ArrayT) :: arr

!    string = trim(stringParameter(param_name)) // " "

!    ncomma1 = 0
!    ncomma2 = index(string, ",") 
!    if (ncomma2 == 0) ncomma2 = index(string, " ")  
       
!    if (ncomma2 <= ncomma1+1) call error ("EmptyString", &
!                                  "Input file name string is empty")

!    nimage = 0
!    do while (ncomma2 > ncomma1+1)
!      nimage = nimage+1
!      imag(nimage) = string(ncomma1+1:ncomma2-1)
!      ncomma1 = ncomma2      
!      ncomma2 = index(string(ncomma1+1:), ",") 
!      if (ncomma2 == 0) ncomma2 = index(string(ncomma1+1:), " ")  
!      ncomma2 = ncomma1 + ncomma2
!    end do  

    nimage = parameterCount(param_name) 
    if (nimage == 0) then
      call error("MissingParameter", "Missing input file name")
    end if
 
    if (first) then 
    
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"EXP_ID")) then
          expid = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expid = stringAttribute(arr,"EXP_ID")
          else
            expid = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if
          
        instrume(i) = trim(teles) // trim(instru) // trim(expid)

!        if (index(instru,"PN") > 0 .or. index(instru,"pn") > 0) then 
!          expip(i) = 1
!        else if (index(instru,"MOS1") > 0 .or. index(instru,"mos1") > 0) then
!          expip(i) = 2
!        else if (index(instru,"MOS2") > 0 .or. index(instru,"mos2") > 0) then
!          expip(i) = 3
!        else
!          expip(i) = 4
!          call warning("WrongInst","Unknown instrument")
!        end if
 
        if (hasAttribute(set,"BAND")) then
          band(i) = int32Attribute(set, "BAND")
        else
          band(i) = 7
        end if
        call release(set)
      end do

      num_instr = 0
      double = .false.    
      ip = -1
      do i=1,nimage        
        if (.not. double(i)) num_instr = num_instr+1
!         print *, i, num_instr
        ib = 1 
        do while (ip(num_instr,ib) /= -1)
          ib = ib+1
        end do
        ip(num_instr,ib) = i
        num_image(num_instr) = ib
        do j=i+1,nimage
          if (instrume(i) == instrume(j)) double(j) = .true.
        end do
      end do

      do i=1,num_instr
        ipoint=ip(i,1) 
        if (index(instrume(ipoint),"PN") > 0 .or. &
            index(instrume(ipoint),"pn") > 0) then 
          expip(i) = 1
        else if (index(instrume(ipoint),"MOS1") > 0 .or. &
                 index(instrume(ipoint),"mos1") > 0) then
          expip(i) = 2
        else if (index(instrume(ipoint),"MOS2") > 0 .or. &
                 index(instrume(ipoint),"mos2") > 0) then
          expip(i) = 3
        else
          expip(i) = 4
          call error("WrongInst","Unknown instrument")
        end if
      end do
        
    else

      if (param_name(1:3) == "det") then
        if (nimage /= num_instr) then
           call error("FileMismatch", &
             "# detector masks /= # instruments")
        end if
      else 
        if (nimage /= nim) then
          call error("FileMismatch","Inconsistent number of input images")
        end if
      end if
 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        ins = trim(teles) // trim(instru) //trim(expid)

        if (hasAttribute(set,"BAND")) then
          ban = int32Attribute(set, "BAND")
        else
          ban = 7
        end if 

        call release(set)

        if (ban < 6 .and. (ins /= instrume(i) .or. ban /= band(i)) ) &
          call error ("FileMismatch", "Inconsistent instruments or bands")  
!        if (ban > 4) then
!          if (ins /= instrume(ip(i,1)) ) &
!            call error ("FileMismatch", &
!                         "Inconsistent instruments or bands")  
!        end if
      end do

    end if

    nim = ip(num_instr,num_image(num_instr))

  end subroutine get_files 

!!!!! OBSOLETE
!!  !---------------------------------------------------------------------
!!  subroutine co_image (fname, start, step, npix, exposure, xcen, ycen, &
!!                       ra_pnt, dec_pnt, pa_pnt, &
!!                       crpix1, crpix2, crval1, crval2, &
!!                       crdelt1,crdelt2,crota2,xoff,yoff,inst,ipoint, nim)
!!  !---------------------------------------------------------------------
!!
!!    character(len=1024), intent(in) :: fname
!!    integer, intent(in) :: inst, ipoint, nim 
!!    real(kind=single), dimension(3), intent(in) :: xoff, yoff
!!    integer, dimension(2), intent(out) :: npix
!!    real(kind=double), dimension(2), intent(out) :: start, step 
!!    real(kind=double), intent(out) :: ra_pnt, dec_pnt, pa_pnt, &
!!                       crpix1, crpix2, crval1, crval2, &
!!                       crdelt1, crdelt2, crota2
!!    real(kind=double) ::cos_pa, sin_pa, rr, dr, r0r, d0r, &
!!                       dra, ddec, xr, yr, del, &
!!                       crpix1old,crpix2old,crval1old,crval2old, &
!!                       crdelt1old, crdelt2old
!!
!!    real(kind=single) :: exposure, xcen, ycen, r4huge
!!    real(kind=double) :: r8huge
!!    integer, dimension(3) :: ixoff, iyoff
!!    integer :: i, ix, iy, iix, iiy, atype
!!    logical, save :: first = .true.
!!
!!    integer(kind=int32), dimension(:,:), pointer :: i4arr
!!    integer(kind=int16), dimension(:,:), pointer :: i2arr
!!    integer(kind=int8), dimension(:,:), pointer :: i1arr
!!    real(kind=single), dimension(:,:), pointer :: r4arr
!!    real(kind=double), dimension(:,:), pointer :: r8arr
!!    integer, dimension(:), pointer :: shape
!!
!!    type(datasett) :: set
!!    type(arrayt) :: arr
!!
!!    r4huge=huge(1)
!!    r8huge=huge(1)
!!
!!    del = 1d-5
!!
!!    set = dataset (fname, read)
!!    arr = array(set, "PRIMARY", read)
!!
!!    shape => dimensions(arr)
!!    npix(1) = shape(0)
!!    npix(2) = shape(1)
!!
!!    if (first) then
!!      allocate (iimage(nim,npix(1),npix(2)))
!!      first=.false.
!!    end if
!!
!!    do i=1,3
!!      ixoff(i)=nint(xoff(i))
!!      iyoff(i)=nint(yoff(i))
!!    end do
!!
!!
!!    atype = arrayDataType(arr)
!!     if (atype == Integer8) then
!!       i1arr => int8array2data(arr)
!!       do iy = 1,npix(2)
!!         do ix = 1,npix(1)
!!           iix=ix-1-ixoff(inst)
!!           iiy=iy-1-iyoff(inst)
!!           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
!!               iiy >= 0 .and. iiy <=npix(2)-1) &
!!             iimage(ipoint,ix,iy) = i1arr(iix,iiy)
!!         end do
!!       end do 
!!     else if (atype == Integer16) then
!!       i2arr => int16array2data(arr)
!!       do iy = 1,npix(2)
!!         do ix = 1,npix(1)
!!           iix=ix-1-ixoff(inst)
!!           iiy=iy-1-iyoff(inst)
!!           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
!!               iiy >= 0 .and. iiy <=npix(2)-1) &
!!           iimage(ipoint,ix,iy) = i2arr(iix,iiy)
!!         end do
!!       end do 
!!     else if (atype == Integer32) then 
!!       i4arr => int32array2data(arr)
!!       do iy = 1,npix(2)
!!         do ix = 1,npix(1)
!!           iix=ix-1-ixoff(inst)
!!           iiy=iy-1-iyoff(inst)
!!           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
!!               iiy >= 0 .and. iiy <=npix(2)-1) &
!!           iimage(ipoint,ix,iy) = i4arr(iix,iiy)
!!         end do
!!       end do 
!!     else if (atype == Real32) then
!!       r4arr => Real32array2data(arr)
!!       do iy = 1,npix(2)
!!         do ix = 1,npix(1)
!!           iix=ix-1-ixoff(inst)
!!           iiy=iy-1-iyoff(inst)
!!           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
!!               iiy >= 0 .and. iiy <=npix(2)-1) &
!!           iimage(ipoint,ix,iy) = nint(min(r4arr(iix,iiy),r4huge))
!!         end do
!!       end do 
!!     else if (atype == Real64) then
!!       r8arr => Real64array2data(arr)
!!       do iy = 1,npix(2)
!!         do ix = 1,npix(1)
!!           iix=ix-1-ixoff(inst)
!!           iiy=iy-1-iyoff(inst)
!!           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
!!               iiy >= 0 .and. iiy <=npix(2)-1) &
!!           iimage(ipoint,ix,iy) = nint(min(r8arr(iix,iiy),r8huge))
!!         end do
!!       end do 
!!     else
!!       call error("WrongType","Input image has wrong type")
!!     end if
!!
!!    start(1) = 1d0
!!    start(2) = 1d0
!!    step(1) = 1d0
!!    step(2) = 1d0
!!
!!    if (hasAttribute(set,"EXPOSURE")) then
!!      exposure = real32Attribute(set,"EXPOSURE")
!!    else if (hasAttribute(arr,"EXPOSURE")) then
!!      exposure = real32Attribute(arr,"EXPOSURE")
!!    else
!!      exposure = 1.0
!!      call warning("MissingAttribute", &
!!                   "No exposure keyword present: set to 1.0")
!!    end if 
!!
!!    if (hasAttribute(set,"RA_PNT")) then
!!      ra_pnt = real64Attribute(set,"RA_PNT")
!!    else  
!!      if (hasAttribute(arr,"RA_PNT")) then
!!        ra_pnt = real64Attribute(arr,"RA_PNT")
!!      else
!!        ra_pnt = 0d0
!!        call warning("MissingAttribute","Keyword RA_PNT is missing")
!!      end if
!!    end if
!!
!!    if (hasAttribute(set,"DEC_PNT")) then
!!      dec_pnt = real64Attribute(set,"DEC_PNT")
!!    else  
!!      if (hasAttribute(arr,"DEC_PNT")) then
!!        dec_pnt = real64Attribute(arr,"DEC_PNT")
!!      else
!!        dec_pnt = 0d0
!!        call warning("MissingAttribute","Keyword DEC_PNT is missing")
!!      end if
!!    end if
!!
!!    if (hasAttribute(set,"PA_PNT")) then
!!      pa_pnt = real64Attribute(set,"PA_PNT")
!!    else  
!!      if (hasAttribute(arr,"PA_PNT")) then
!!        pa_pnt = real64Attribute(arr,"PA_PNT")
!!      else
!!        pa_pnt = 0d0
!!        call warning("MissingAttribute","Keyword PA_PNT is missing")
!!      end if
!!    end if
!!
!!    crpix1old=crpix1
!!    if (hasAttribute(arr,"CRPIX1")) then
!!      crpix1 = real64Attribute(arr,"CRPIX1")
!!      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CRPIX1")) then
!!        crpix1 = real64Attribute(set,"CRPIX1")
!!      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CRPIX1 is missing")
!!      end if
!!    end if
!!
!!    crpix2old=crpix2
!!    if (hasAttribute(arr,"CRPIX2")) then
!!      crpix2 = real64Attribute(arr,"CRPIX2")
!!      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CRPIX2")) then
!!        crpix2 = real64Attribute(set,"CRPIX2")
!!        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
!!           call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CRPIX2 is missing")
!!      end if 
!!    end if
!!
!!    crval1old=crval1
!!    if (hasAttribute(arr,"CRVAL1")) then
!!      crval1 = real64Attribute(arr,"CRVAL1")
!!      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CRVAL1")) then
!!        crval1 = real64Attribute(set,"CRVAL1")
!!      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CRVAL1 is missing")
!!      end if
!!    end if
!!
!!    crval2old=crval2
!!    if (hasAttribute(arr,"CRVAL2")) then
!!      crval2 = real64Attribute(arr,"CRVAL2")
!!      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CRVAL2")) then
!!        crval2 = real64Attribute(set,"CRVAL2")
!!      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CRVAL2 is missing")
!!      end if
!!    end if
!!
!!
!!    crdelt1old=crdelt1
!!    if (hasAttribute(arr,"CDELT1")) then
!!      crdelt1 = real64Attribute(arr,"CDELT1")
!!      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CDELT1")) then
!!        crdelt1 = real64Attribute(set,"CDELT1")
!!        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
!!           call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CRDELT1 is missing")
!!      end if
!!    end if
!!
!!    crdelt2old=crdelt2
!!    if (hasAttribute(arr,"CDELT2")) then
!!      crdelt2 = real64Attribute(arr,"CDELT2")
!!      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
!!         call error("FileMismatch","Inconsistent WCS keywords")
!!    else
!!      if (hasAttribute(set,"CDELT2")) then
!!        crdelt2 = real64Attribute(set,"CDELT2")
!!        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
!!           call error("FileMismatch","Inconsistent WCS keywords")
!!      else
!!        call warning("MissingAttribute","Keyword CDELT2 is missing")
!!      end if
!!    end if
!! 
!!    if (hasAttribute(arr,"CROTA2")) then
!!       crota2 = real64Attribute(arr,"CROTA2")
!!    else
!!      if (hasAttribute(set,"CROTA2")) then
!!        crota2 = real64Attribute(set,"CROTA2")
!!      else
!!        crota2 = 0d0
!!        call warning("MissingAttribute","Keyword CROTA2 is missing")
!!      end if
!!    end if
!!
!!    cos_pa = dcos(crota2 * pi/180d0)
!!    sin_pa = dsin(crota2 * pi/180d0)
!!    rr = ra_pnt * pi/180d0
!!    dr = dec_pnt * pi/180d0  
!!    r0r = crval1 * pi/180d0
!!    d0r = crval2 * pi/180d0
!!    call spher2(rr, dr, r0r, d0r, dra, ddec)
!!    xr =   dra * cos_pa + ddec * sin_pa ! Vorzeichen ??
!!    yr = - dra * sin_pa + ddec * cos_pa ! 
!!    xcen = xr * 180d0/pi / crdelt1 + crpix1
!!    ycen = yr * 180d0/pi / crdelt2 + crpix2
!!
!!    call release(arr)
!!    call release(set)
!!
!!  end subroutine co_image

  !---------------------------------------------------------------------
  subroutine co_bgima (fname, npix, &
                       crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                       xoff, yoff, inst, ipoint, nim, stat)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    character(len=10), intent(in) :: stat

    integer, dimension(3) :: ixoff, iyoff
    integer :: ix, iy, iix, iiy, i, l
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del
    integer, dimension(:), pointer :: shape

    type(datasett) :: set
    type(arrayt) :: arr

    type(SpacecraftAttitudeType) :: scAttitude
    type(SpacecraftAttitudeType) :: instrAttitude
    real(kind=double)            :: timeStamp
    integer(int32) :: instidtmp

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
       allocate (bg_image(npix(1),npix(2),nim))
       ! get wcs information from first image
       if (stat == "deltac") call readWcs(set, arr, wcs)
       first=.false.
    end if

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    do iy = 1,npix(2)
      do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1) &
        bg_image(ix,iy,ipoint) = r4arr(iix,iiy)
        l = l+1
      end do
    end do 
 
    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if


    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    ! attitude information
    if (stat == "deltac") then
       call getCalInfo(set, instidtmp, timeStamp, scAttitude, instrAttitude)
       timestamps(ipoint) = timeStamp
       attitudes(ipoint) = scAttitude
    end if

    call release(arr)
    call release(set)

  end subroutine co_bgima

  !---------------------------------------------------------------------
  subroutine co_expima (fname, npix, &
                        crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                        xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 

    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del
    integer, dimension(:), pointer :: shape

    type(datasett) :: set
    type(arrayt) :: arr

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

! this is preliminary: -------------
!    xcen1 = npix(1)/2
!    ycen1 = npix(2)/2
!-----------------------------------

    if (first) then
!     print *,'allocate exp image'
      allocate (exp_image(npix(1),npix(2),nim))
      first=.false.
    end if

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    do iy = 1,npix(2)
      do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
        exp_image(ix,iy,ipoint) = r4arr(iix,iiy)
      end do
    end do 

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    call release(arr)
    call release(set)

  end subroutine co_expima

  !---------------------------------------------------------------------
  subroutine co_mask (fname, npix, &
                      crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                      xoff, yoff, inst, ipoint, iinst)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint , iinst
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 

    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy
    logical, save :: first = .true.

    integer, dimension(:,:), pointer :: i4arr
    integer, dimension(:), pointer :: shape
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    integer, allocatable :: isumx(:), isumy(:)

    type(datasett) :: set
    type(arrayt) :: arr

    del = 1d-5

    set = dataset (fname, read)

    !! The mask extension should be by default in an extension called
    !! MASK, if not then look in the PRIMARY array ! RDS
    if ( hasBlock( set, "MASK" ) ) then
        arr = array(set, "MASK", read)
    else
        arr = array(set, "PRIMARY", read)
    endif

    i4arr => int32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) allocate (mask_image(npix(1),npix(2),nim))

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    do iy = 1,npix(2)
      do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1)  then
          if (i4arr(iix,iiy) == 0) then
            mask_image(ix,iy,ipoint) = .false.
          else
            mask_image(ix,iy,ipoint) = .true.
          end if
        end if
      end do
    end do 


    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if


    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        crval1 = 0d0
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMismatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMismatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    if (hasAttribute(arr,"PA_PNT")) then
      pa_pnt(iinst) = real64Attribute(arr,"PA_PNT")
    else  
      if (hasAttribute(set,"PA_PNT")) then
        pa_pnt(iinst) = real64Attribute(set,"PA_PNT")
      else
        pa_pnt(iinst) = 0d0
        call warning("MissingAttribute","Keyword PA_PNT is missing")
      end if
    end if
    
    ! determine coordinates of the exposed image (mask) area:
    ! (so far only used to limit the loops; could be also used for array sizes)
    allocate(isumx(npix(1)))
    allocate(isumy(npix(2)))
    if (first) then    !!! keep the results from previous calls
       imxmin=npix(1)
       imymin=npix(2)
       imxmax=1
       imymax=1
    endif

    ! collapse
    isumx = sum(i4arr,2)
    isumy = sum(i4arr,1)

    ! - first column with a non-zero entry (i.e.: along x),  bottom-up
    do ix=1,imxmin-1
       if (isumx(ix)>0) then
          imxmin=ix
          exit
       end if
    end do
    ! - first row    with a non-zero entry (i.e.: along y),  bottom-up
    do iy=1,imymin-1
       if (isumy(iy)>0) then
          imymin=iy
          exit
       end if
    end do
    ! - last  column with a non-zero entry (i.e.: along x),  top-down
    do ix=npix(1),imxmax+1,-1
       if (isumx(ix)>0) then
          imxmax=ix
          exit
       end if
    end do
    ! - last  row    with a non-zero entry (i.e.: along y),  top-down
    do iy=npix(2),imymax+1,-1
       if (isumy(iy)>0) then
          imymax=iy
          exit
       end if
    end do

    ! add some margin and round
    imxmin = max( (imxmin-16)/16*16, 1 )
    imymin = max( (imymin-16)/16*16, 1 )
    imxmax = min( (imxmax+16)/16*16-1, npix(1) )
    imymax = min( (imymax+16)/16*16-1, npix(2) )
    imusex = imxmax-imxmin+1
    imusey = imymax-imymin+1
    
    ! done
    call release(arr)
    call release(set)
    first=.false.

  end subroutine co_mask

end module edetect_io_sens







