C ESA (C) 2000-2018 
C
C This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
C
C    SAS is free software: you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation, either version 3 of the License, or
C    (at your option) any later version.
C
C    SAS is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
      FUNCTION GAMMLN (XX)
C
CC  returns natural logarithm of the Gamma function for XX>=1
C
C************************ FFORM VERSION 1.1 ************  7-NOV-89 16:16
C
CA  author : S.Doebereiner date: 11-SEP-1996
C
CT  status: not tested
C
C   general description 
CG  Returns natural logarithm of the Gamma function for XX>=1
CG  Routine GAMMLN from Press et al., Numerical Recipes, 6.1, p.157
CG
C
C   call_var.          type I/O description 
CP  XX                  R4  I   argument
CP  GAMMLN              R4    O LN(Gamma(XX))
C
C   include_block name          description 
CI  RCOMMON:CGENL.CMN  general parameter common block
CI                              
C
C   routines_called    type     description 
CR  WRFLAG              SR      writes to FLAG output stream
CR  WRERR               SR      write error message
C
C   extensions/system calls     description 
CX 
C
C***********************************************************************
C
C   variables   meaning
C
C
      IMPLICIT NONE
c     INCLUDE 'include/cgenl.cmn'
C
      CHARACTER*8   RTNAME
      REAL*4        GAMMLN,XX 
      REAL*8        COF(6),STP,HALF,ONE,FPF,X,TMP,SER
      INTEGER       J,IERR
C
      SAVE
C
      DATA          RTNAME /'GAMMLN'/
      DATA COF,STP/76.18009173D0,-86.50532033D0,24.01409822D0,
     *             -1.231739516D0,.120858003D-2,
     *             -.536382D-5,2.50662827465D0/
      DATA HALF,ONE,FPF /0.5D0,1.0D0,5.5D0/
               
C
      IERR = 0
      GAMMLN = 0.
C
      IF (XX.LE.0.0) THEN
        IERR = 1
c        CALL WRERR(RTNAME,IERR,'argument <=0 not allowed ') 
      ENDIF
C
      IF (IERR.EQ.0) THEN 
C
C if XX<1 then use reflection formula (see Press et al., eq. 6.1.4)
C     
        IF (XX.LT.1) THEN
          X = XX + 1.D0 - ONE
        ELSE
          X = XX - ONE
        ENDIF       
C
        TMP=X+FPF
        TMP=(X+HALF)*LOG(TMP)-TMP
        SER=ONE
        DO 111 J=1,6
          X=X+ONE
          SER=SER+COF(J)/X
  111   CONTINUE
C
C reflection formula for XX<1
C
        IF (XX.LT.1.) THEN
          GAMMLN=(TMP+LOG(STP*SER))/XX
        ELSE
          GAMMLN=TMP+LOG(STP*SER)        
        ENDIF 
C
      ENDIF
C
      RETURN
      END
